#!/bin/bash
export LD_LIBRARY_PATH=/usr/local/mysql/lib/mysql

export SOURCEROOT=$1
export JOB_NAME=$2
export BUILD_ID=$3

export JAVA_HOME=/usr/j2sdk1.4.1_02
export ANT_HOME=/usr/local/apache-ant-1.6.1
export PATH=$JAVA_HOME/bin:$ANT_HOME/bin:/usr/local/bin:$PATH
export DIST_DIR=/mnt/dev/builds/$JOB_NAME/$BUILD_ID
export BUILD_NUM=$(cat $SOURCEROOT/$JOB_NAME.build)

mkdir -p "$DIST_DIR"

cat > $DIST_DIR/build.properties <<-EOF
	build.num=$BUILD_NUM
	buildserver.dir=$(pwd)
	dist.dir=$DIST_DIR
	jdk=$JAVA_HOME
	job.title=$JOB_NAME
EOF

export ANT_ARGS="-propertyfile $DIST_DIR/build.properties"

P4OPTS="-u builduser -P password -p perforce_server -c jobname_bldserver"

echo ""
echo "DATE        = $(date)"
echo "PWD         = $(pwd)"
echo "SOURCEROOT  = $SOURCEROOT"
echo "DIST_DIR    = $DIST_DIR"
echo "JOB_NAME    = $JOB_NAME"
echo "BUILD ID    = $BUILD_ID"
echo "BUILD NUM   = $BUILD_NUM"
echo "JAVA_HOME   = $JAVA_HOME"
echo "ANT_HOME    = $ANT_HOME"
echo "ANT_ARGS    = $ANT_ARGS"
echo ""

# ----------------------------------------------------------------------------
ant build.release || exit 1

rm -rf $DIST_DIR/_repo \
       $DIST_DIR/_task \

# ----------------------------------------------------------------------------
export BLDSVRDIR=$HOME/buildserver/server
export SHARED_DIR=$BLDSVRDIR/jobs/shared
perl -I$BLDSVRDIR/lib $SHARED_DIR/buildfuncs.pl jdiff \
        $SOURCEROOT/jarfile.jar \
        $SOURCEROOT/jarfile/common.jar

if [ $? = 1 ]
then
    #
    # Check it in.
    #
    (
        cd $SOURCEROOT/cmsbinaries/reference

        echo ""
        echo "------------------------------------------------"
        echo "  Copy sct_common.jar to reference dir"
        echo "------------------------------------------------"
        echo ""

        p4 $P4OPTS edit common.jar

        cp $SOURCEROOT/sctcommon/dist/sct_common.jar \
           $SOURCEROOT/cmsbinaries/reference

        comment="update from build $BUILD_ID of $JOB_NAME"

        p4 $P4OPTS change -o |
        sed "s/<enter description here>/$comment/g" |
        p4 $P4OPTS submit -i
    )
fi

exit 0
